#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <sys/types.h>		// open()
#include <unistd.h>		// read() / write()
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <stdio.h>		// FILE
#include <stdlib.h>		// atol()
#include <string.h>		// strstr()
#include <errno.h>          	// EIO, E...

#include <linux/adwin.h>
#include "dsp.h"

#include "libadwin/errno.h"

extern int errno;

#define VARDUMP(i)	printf("  %s: %ld\n", #i, i)


#define write_save(dev, data, len)					\
	if (write(dev, (char *)(data), (len)) != (len)) {	 	\
		errno = ETIMEDOUT;					\
		adwin_errno = k_timeout_to_link;				\
		return -1;						\
	}

#define write_save_and_free(dev, data, len, ptr_to_free) 		\
	if (write(dev, (char *)(data), (len)) != (len))	{	 	\
		free(ptr_to_free);					\
		errno = ETIMEDOUT;					\
		adwin_errno =  k_timeout_to_link;			\
		return -1;						\
	}

#define read_save(dev, data, len) 					\
	if (read(dev, (char *)(data), (len)) != (len)) {		\
		errno = ETIMEDOUT;					\
		adwin_errno = k_timeout_from_link;			\
		return -1;						\
	}

#define read_save_and_free(dev, data, len, ptr_to_free) \
	if (read(dev, (char *)(data), (len)) != (len)) {		\
		free(ptr_to_free);					\
		errno = ETIMEDOUT;					\
		adwin_errno = k_timeout_from_link;			\
		return -1;						\
	}


static int __adbload(adwin_dev_t *dev, int dsp, FILE **bin);

static int __opcodes_to_dsp(int fd, const unsigned char *bytes, int cnt) {
	int i;
	unsigned char buffer[6] = {0};
	unsigned char *ptr = (unsigned char*)bytes;

	for (i = 0; i < cnt; i += 6) {
		buffer[0] = ptr[4];
		buffer[1] = ptr[5];
		buffer[2] = ptr[2];
		buffer[3] = ptr[3];
		buffer[4] = ptr[0];
		buffer[5] = ptr[1];
		if (ioctl(fd, ADIOSOPCODE, buffer) == -1) {
			errno = ETIMEDOUT;
			adwin_errno =  k_timeout_to_link;
			return -1;
		}

		ptr += 6;
	}
	return 0;
}


long DSP_Boot(adwin_dev_t *dev, int fd, const char *dateiname, long iboardsize) {
	FILE *btl;

	unsigned char *bootcode;
	long anzahl, anzahl_proc;
	long proc_nr, proc_groesse;
	int i;

	ioctl(fd, ADIORESETLINK);
	ioctl(fd, ADIOSKDEVPROC, DP_DSP);

	if((btl = fopen(dateiname, "rb")) == NULL) {
		adwin_errno = -errno;
		return -1;
	}

	/* Bootcode */
	fread(&anzahl, 4, 1, btl);

	if((bootcode = (unsigned char *)malloc(anzahl)) == NULL)
		return -1;

	fread(bootcode, 1, anzahl, btl);

	if (__opcodes_to_dsp(fd, bootcode, anzahl) < 0) {
		free(bootcode);
		return -1;
	}
	free(bootcode);

	usleep(100);

	if (DSP_Test_Version(dev, fd) < 0) {
		fclose(btl);
		return -1;
	}

	fread(&anzahl_proc, sizeof(anzahl_proc), 1, btl);
	for (i = 0; i < anzahl_proc; i++) {
		fread(&proc_groesse, sizeof(proc_groesse), 1, btl);
		fread(&proc_nr, sizeof(proc_nr), 1, btl);

		if (__adbload(dev, fd, &btl) < 0)
			return -1;
	}

	fclose(btl);

	/* Prozess 15 starten */
	DSP_Start_Process(dev, fd, 129);
	errno = 0;
	return 8000;
}

processor_type_t DSP_Processor_Type(adwin_dev_t *dev, int fd, processor_type_t *processor) {
        // not yet implemented
        *processor = -1;
	return -1;
}

/*
long DSP_AD_Memory(adwin_dev_t *dev, int fd) {
	int32_t buffer[3] = { 5, 0, 0 };
	long res = 0;

	write_save(fd, buffer, 12);
	read_save(fd, &res, 4);

	return res;
}
*/
int DSP_Free_Mem(adwin_dev_t *dev, int fd, int Mem_Spec, int32_t *dest) {
	int32_t buffer[3] = { 12, Mem_Spec, 0 };

	write_save(fd, buffer, 12);
	read_save(fd, dest, 4);

	return 0;
}


int DSP_Get_Memory_dsp(adwin_dev_t *dev, int fd, int base_addr,
				int anzahl, int32_t *dest) {
	int32_t buffer[3] = { 1000000, base_addr, anzahl };

	write_save(fd, buffer, 12);
	read_save(fd, dest, anzahl * 4);

	return 0;
}

int DSP_Test_Version(adwin_dev_t *dev, int fd) {
	int32_t buffer[3] = { 255, 255, 255 };
	long tmp[4] = { 0 };

	write_save(fd, buffer, 12);
	read_save(fd, tmp, 16);

	if (tmp[0] != 118) {
		/* Schwerer Fehler! (Bitfehler) */
		errno = EIO;
		adwin_errno = k_wrong_version;
		return -1;
	}

	switch(tmp[1]) {
		case 1000:	dev->dev_processor = DP_DSP; 		break;
		case 1100:	dev->dev_processor = DP_21160; 		break;
		default:	dev->dev_processor = DP_UNKNOWN; 	break;
	}
	
	ioctl(fd, ADIOSKDEVPROC, dev->dev_processor);

	return 0;
}


int DSP_Load_Process(adwin_dev_t *dev, int fd, const char *dateiname) {
	FILE *bin;

	if ((bin = fopen(dateiname, "rb")) == NULL)
		return -1;

	if (__adbload(dev, fd, &bin) < 0)
		return -1;

	fclose(bin);
	return 0;
}


static int __adbload(adwin_dev_t *dev, int fd, FILE **bin) {
	int32_t buffer[6], tmp;
	long Version,anzahlpatches;
	long anzahl, delay, anzahlloop;
	int j;

	char ProcID;
	long groesse;
	long stacklaenge;

	unsigned char *processcode;

	fread(&Version, 4, 1, *bin);
	fread(&Version, 4, 1, *bin);

	fread(&ProcID, 1, 1, *bin);
	fread(&stacklaenge, 4, 1, *bin);
	buffer[0] = 4;
	buffer[1] = ProcID;
	buffer[2] = stacklaenge;

	write_save(fd, buffer, 12);
	read_save(fd, &tmp, 4);

	fread(&groesse, 4, 1, *bin);
	buffer[0] = groesse;
	fread(buffer + 1, 4, 5, *bin);

	write_save(fd, buffer, 24);
	read_save(fd, &tmp, 4);

	fread(buffer, 4, 4, *bin);
	write_save(fd, buffer, 16);
	read_save(fd, &tmp, 4);

	fread(buffer, 4, 4, *bin);
	anzahlpatches = buffer[3];
	write_save(fd, buffer, 16);
	read_save(fd, &tmp, 4);

	// nun Code
	processcode = (unsigned char *)malloc(6 * groesse);
	if (processcode == NULL)
		return -1;

	fread(processcode, 6, groesse, *bin);
	fread(&anzahl, 4, 1, *bin);

	for (j = 0; j < anzahlpatches; j++) {
		long value, position, addresse;

		fread(&value, 4, 1, *bin);
		write_save(fd, &value, 4);
		read_save(fd, &value, 4);
		fread(&position, 4, 1, *bin);
		addresse = 6 * position;

		processcode[addresse + 1] = value & 0xff;
		processcode[addresse] = (value >> 8) & 0xff;
		processcode[addresse + 2] = (value >> 16) & 0xff;
	}

	__opcodes_to_dsp(fd, processcode, 6 * groesse);
	read_save(fd, &tmp, 4);

	free(processcode);

	/* CreateData */
	fread(&anzahl, 4, 1, *bin);
	for (j = 0; j < anzahl; j++) {
		buffer[0] = 10;
		fread(buffer + 1, 4, 2, *bin);
		write_save(fd, buffer, 12);
		read_save(fd, &tmp, 4);

		fread(buffer, 4, 3, *bin);
		write_save(fd, buffer, 12);
		read_save(fd, &tmp, 4);
	}

	fread(&delay, 4, 1, *bin);
	fread(&anzahlloop, 4, 1, *bin);

	if ((ProcID >= 1) && (ProcID <= 10)) {
		if (DSP_Set_Par(dev, fd, 910 + ProcID, delay) < 0)
			return -1;
		if (DSP_Set_Par(dev, fd, 920 + ProcID, anzahlloop) < 0)
			return -1;
	}

	return 0;
}

int DSP_Start_Process (adwin_dev_t *dev, int fd, int processID) {
	int32_t buffer[3];
	long res = 0;
	long tmp;

	if (processID < 100) {
		switch (processID) {
			case 8:
				processID = 123;
				break;

			case 1:
			{
				int32_t lh, maske, i;

				processID = 121;

				/* Anhalten, implizites starten s.u. */
				DSP_Stop_Process(dev, fd, processID);
				/* Memaske */
				DSP_Get_Par(dev, fd, 1081, &maske);
				/* Anzahl Werte */
				DSP_Get_Par(dev, fd, 1082, &lh);
				/* Gre Array setzen */
				DSP_Set_Par(dev, fd, 1089, lh + 10);

				for(i = 0; i < 16; i++) {
					if (maske & (1 << i)){
						/* Create Data */
						buffer[0] = 10;
						buffer[1] = 201;
						buffer[2] = 3;
						write_save(fd, buffer, 12);
						read_save(fd, &tmp, 4);

						buffer[0] = lh + 10;
						buffer[1] = 2;
						buffer[2] = 0;
						write_save(fd, buffer, 12);
						read_save(fd, &tmp, 4);
					}
				}
			}
			break;
		}
	}

	buffer[0] = 2;
	buffer[1] = (processID - 101)/2 + 1;
	buffer[2] = 0;

	write_save(fd, buffer, 12);
	read_save(fd, &res, 4);

	return res;
}

int DSP_Stop_Process (adwin_dev_t *dev, int fd, int processID) {
	int32_t buffer[3] = { 3, (processID - 101)/2 + 1, 0 };
	long tmp;

	write_save(fd, buffer, 12);
	read_save(fd, &tmp, 4);

	return 0;
}

int  DSP_Clear_Process(adwin_dev_t *dev, int fd, int processID) {
	int32_t buffer[3] = { 14, (processID - 101)/2 + 1, 0 };
	long tmp;

	write_save(fd, buffer, 12);
	read_save(fd, &tmp, 4);

	return 0;
}

int DSP_Get_Par (adwin_dev_t *dev, int fd, int parID, int32_t *dest) {
	int32_t buffer[3] = { 1, parID, 0 };

	write_save(fd, buffer, 12);
	read_save(fd, dest, 4);

	return 0;
}

int DSP_Set_Par (adwin_dev_t *dev, int fd, int parID, int32_t value) {
	int32_t buffer[3] = { 6, parID, value };
	long res = 0;

	write_save(fd, buffer, 12);
	read_save(fd, &res, 4);

	return res;
}

int DSP_Get_FPar (adwin_dev_t *dev, int fd, int parID, float *dest) {
	int32_t buffer[3] = { 1, parID, 0 };

	write_save(fd, buffer, 12);
	read_save(fd, dest, 4);

	return 0;
}

int DSP_Set_FPar (adwin_dev_t *dev, int fd, int parID, float value) {
	long *lptr;
	int32_t buffer[3];

	lptr = (void*)&value;
	buffer[0] = 6;
	buffer[1] = parID;
	buffer[2] = *lptr;
	long res = 0;

	write_save(fd, buffer, 12);
	read_save(fd, &res, 4);

	return res;
}

int DSP_Get_Par_All(adwin_dev_t *dev, int fd, int first,
		    int cnt, int32_t *dest) {
	int32_t buffer[3] = { 13, first, cnt };

	write_save(fd, buffer, 12);
	read_save(fd, dest, cnt * 4);

	return 0;
}

int DSP_Get_FPar_All(adwin_dev_t *dev, int fd, int first,
		     int cnt, float *dest) {
	int32_t buffer[3] = { 13, first, cnt };

	write_save(fd, buffer, 12);
	read_save(fd, dest, cnt * 4);

	return 0;
}

int DSP_Workload (adwin_dev_t *dev, int fd, int priority, int32_t *dest) {
	int32_t buffer[3] = { 20, 0, 0 };
	long zeit = 0, wert = 100;

	write_save(fd, buffer, 12);
	read_save(fd, &wert, 4);
	read_save(fd, &zeit, 4);

	if (zeit != 0) {
		float help = (float)wert / (float) zeit;
		wert = (unsigned short)(100. * ((0.25 - help) / 0.25)) ;
	}

	*dest = wert;
	return 0;
}

int DSP_Data_Length(adwin_dev_t *dev, int fd, long DataNo) {
	int32_t buffer[3] = { 11, DataNo, 0 };
	long res;

	write_save(fd, buffer, 12);
	read_save(fd, &res, 4);

	return (res > 0) ? res - 1 : 0;
}

int DSP_Clear_Data(adwin_dev_t *dev, int fd, int Datanr) {
	int32_t buffer[3] = { 9, Datanr, 0 };
	long tmp;

	write_save(fd, buffer, 12);
	read_save(fd, &tmp, 4);

	return 0;
}

/*
int  DSP_CreateData(adwin_dev_t *dev, int fd, int datanr, int type, long groesse, int ort, int art) {
	int32_t buffer[3] = { 10, datanr, type };
	long tmp;

	write_save(fd, buffer, 12);
	read_save(fd, &tmp, 4);

	buffer[0] = groesse;
	buffer[1] = ort;
	buffer[2] = art;
	write_save(fd, buffer, 12);
	read_save(fd, &tmp, 4);

	return 0;
}
*/
int DSP_GetDataType(adwin_dev_t *dev, int fd, int dataID) {
	int32_t buffer[3] = { 7, dataID, 1 };
	long res;

	write_save(fd, buffer, 12);
	read_save(fd, &res, 4);

	if (res == 0) {
		adwin_errno = k_Data_is_to_small;
		errno = EINVAL;
		return -1;
	}

	/* Fr was das gut ist, ist sich Markus auch nicht mehr ganz sicher */
	/* Siehe "GetData()" ... diese Funktion ist nichts anderes als ein
	 * GetData(dataID, first = 1, count = 1)
	 */
	buffer[0] = 1;
	write_save(fd, buffer, 4);
	read_save(fd, buffer, 4);

	return res;
}

int DSP_GetData(adwin_dev_t *dev, int fd, int dataNo, long startindex, long count, void *dest) {

	int32_t buffer[3] = { 7, dataNo, startindex };
	int data_type;

	write_save(fd, buffer, 12);
	read_save(fd, &data_type, 4);

	if (data_type == 0) {
		adwin_errno = k_Data_is_to_small;
		errno = EINVAL;
		return -1;
	}

	/*
	 * If data has dim n and you ask for n + m elements, no problem there.
	 * you will get the next m elements from memory. Contents: undefined.
	 */
	write_save(fd, &count, 4);
	read_save(fd, dest, count * 4);

	return 0;
}
int DSP_SetData(adwin_dev_t *dev, int fd, int dataNo, long startindex, long count, void *src) {

	int32_t buffer[3] = { 8, dataNo, startindex };
	long data_type, tmp;

	write_save(fd, buffer, 12);
	read_save(fd, &tmp, 4);

	write_save(fd, &count, 4);
	read_save(fd, &data_type, 4);

	if (data_type == 0) {
		adwin_errno = k_Data_is_to_small;
		errno = EINVAL;
		return -1;
	}

	write_save(fd, src, count * 4);
	/* sync */
	read_save(fd, &tmp, 4);

	return 0;
}

int DSP_Fifo_Full(adwin_dev_t *dev, int fd, int FifoNo, int *dest) {
	int32_t buffer[3] = { 23, FifoNo, 0 };

	write_save(fd, buffer, 12);
	read_save(fd, dest, 4);

	return 0;
}

int DSP_Fifo_Empty(adwin_dev_t *dev, int fd, int FifoNo, int *dest) {
	int32_t buffer[3] = { 22, FifoNo, 0 };

	write_save(fd, buffer, 12);
	read_save(fd, dest, 4);

	return 0;
}


int DSP_Fifo_Clear(adwin_dev_t *dev, int fd, int FifoNo) {
	int32_t buffer[3] = { 21, FifoNo, 0 };
	long tmp;

	write_save(fd, buffer, 12);
	read_save(fd, &tmp, 4);

	return 0;
}

int DSP_GetFifo(adwin_dev_t *dev, int fd, int FifoNo, long count, void *dest) {
	int32_t tmp, buffer[3] = { 24, FifoNo, count };
	int fifo_type, fifo_cnt;

	write_save(fd, buffer, 12);
	read_save(fd, &fifo_type, 4);

	/*
	 * implizite Fehlerrckgabe:
	 * wird mehr angefodert als vorhanden, wird auch Fehlerstatus 0
	 * zurckgeliefert
	*/
	if (fifo_type == 0) {
		adwin_errno = k_Fifo_to_less_data;
		errno = EINVAL;
		return -1;
	}

	read_save(fd, &fifo_cnt, 4);
	write_save(fd, &count, 4);

	if (fifo_cnt == count) {
		read_save(fd, dest, count * 4);
	} else {
		read_save(fd, dest, fifo_cnt * 4);
		/*
		 * Daniel:
		 * warum hier nicht: write_save(fd, "count - fifo_cnt", 4);
		 * ???
		 */
		write_save(fd, &count, 4);
		read_save(fd, ((int *)dest) + fifo_cnt, (count - fifo_cnt) * 4);
	}

	write_save(fd, &count, 4);
	read_save(fd, &tmp, 4);

	return 0;
}

int DSP_SetFifo(adwin_dev_t *dev, int fd, int FifoNo, long count, void *src) {
	int32_t buffer[3] = { 25, FifoNo, count };
	long fifo_type, fifo_cnt;
	long tmp;

	write_save(fd, buffer, 12);
	read_save(fd, &fifo_type, 4);

	if (fifo_type == 0) {
		adwin_errno = k_Fifo_is_to_small;
		errno = EINVAL;
		return -1;
	}

	read_save(fd, &fifo_cnt, 4);

	if (fifo_cnt == count) {
		write_save(fd, src, count * 4);
	} else {
		write_save(fd, src, fifo_cnt * 4);
		read_save(fd, &tmp, 4);
		write_save(fd, ((int *)src) + fifo_cnt, (count - fifo_cnt) * 4);
	}

	/* sync */
	read_save(fd, &tmp, 4);

	return 0;
}




