#ifndef __LIBADWIN_BITS_H__
#define __LIBADWIN_BITS_H__

#include <inttypes.h>

// this is the data type given to Get/Set_Data to determine which type
// you have or want in your program.
typedef enum {
	VARIANT_SHORT  = 2, // 16 Bit
	VARIANT_LONG   = 3, // 32 Bit
	VARIANT_FLOAT  = 4, // 32 Bit
	VARIANT_DOUBLE = 5, // 64 Bit
} variant_data_type_t;

struct myVariant {
	// type has type variant_data_type_t 
	uint16_t type;
	unsigned char dummy[6];
	unsigned char value[8];
};
typedef struct myVariant myVariant_t;

// this is the data type given to Get/Set_Data to determine which type
// you have or want in your program.
typedef enum {
	ADWIN_SHORT  = 1, // 16 Bit
	ADWIN_LONG   = 2, // 32 Bit
	ADWIN_FLOAT  = 5, // 32 Bit
	ADWIN_DOUBLE = 6, // 64 Bit
	ADWIN_VARIANT = 20, // 128 Bit, see above
} adwin_data_type_t;

// these are the internal data types used inside the driver.
typedef enum {
	ADWIN_SYSTEM_SHORT   = 2, // Transputer: 16 Bit, T9 and newer: 32 Bit
	ADWIN_SYSTEM_INTEGER = 3, // Wordsize of processor, T2: 16 Bit, others: 32 Bit
	ADWIN_SYSTEM_LONG    = 4, // 32 Bit
	ADWIN_SYSTEM_FLOAT   = 5, // 32 Bit; not supported on T2 and T4
} adwin_system_data_type_t;

typedef enum {
	MEM_ALL = 0,         /* only T2, T4, T5 and T8 */
	MEM_PM_LOCAL = 1,    /* >= T9 */
	MEM_EM_LOCAL = 2,    /* T11 only */
	MEM_DM_LOCAL = 3,    /* >= T9 */
	MEM_DRAM_EXTERN = 4, /* >= T9 */
} mem_spec_t;

typedef enum {
	PROCESSOR_T2 = 2,
	PROCESSOR_T4 = 4,
	PROCESSOR_T5 = 5,
	PROCESSOR_T8 = 8,
	PROCESSOR_T9 = 9,
	PROCESSOR_T10 = 1010,
	PROCESSOR_T11 = 1011,
} processor_type_t;

#if 0
// TB: This is messed up, every define is contained twice in the
// source code. This file would be the better place to use, but almost
// all projects want these defines to be in UDPADwin.h.

/* 
 * Processor constants
 * (These differ from those <linux/adwin.h> defines ...
 * the k_proc defines are the values that are returned to the user
 * the k_proc_net values are the values received/send on the net
 */
#define k_proc_t2               2
#define k_proc_t4               4
#define k_proc_t5               5
#define k_proc_t8               8
#define k_proc_t9               9
#define k_proc_t10           1010
#define k_proc_t11           1011

#define k_proc_net_t9        1000
#define k_proc_net_t10       1010
#define k_proc_net_t11       1011

/*
 * error codes returned from adwin ethernet server
 * (for internal use only)
 */
#define	k_timeout                  5
#define	k_UDPtimeout              -1
#define k_WrongPassword           -5
#define k_TryLater               -10
#define k_DateiFehler            -15
#define k_PaketFehler            -20
#define k_Paketzugross	         -25
#define k_WrongSize              -26
#define k_WrongVersion	         -30
#define k_PaketZuKlein	         -31
#define k_FifoLessData	         -32
#define k_FalscheReihenfolge     -33
#define k_SchonBearbeitet        -40
#define k_RetryUnknown           -41

/*
 * command shortcuts used with adwin ethernet server
 * (for internal use only)
 */
#define k_3plus1		0
#define k_GetPar		1
#define k_Start			2
#define k_Stop			3
#define k_lMemory		5
#define k_SetPar		6
#define k_Clear_Data		9
#define k_GetDataLength		11
#define k_MemoryAll		12
#define k_ClearProcess		14
#define k_ADC			15
#define k_DAC			16
#define k_GetDigin		17
#define k_SetDigout		18
#define k_GetDigout		19
#define k_ClearFifo		21
#define k_GetFifoEmpty		22
#define k_GetFifoCount		23

#define k_LadeBin		4
#define k_GetData		7
#define k_SetData		8
#define k_CreateData		10
#define k_GetParAll		13
#define k_Auslastung		20
#define k_GetFifo		24
#define k_SetFifo		25
#define k_Boot			50
#define k_GetDataTyp		100

#define k_GetDatafast		107

#define k_GetFifo_Retry		124
#define k_SetFifo_Retry		125

#define k_TestVersion		255
#define k_armVersion		1000
#define k_GetDatasmall		207
#define k_GetMem		1000000

#endif // 0 

#endif // __LIBADWIN_BITS_H__
